/*----------------------------------------------------------------------------

linalg.h - A very cheesy linear algebra library.

10-25-99 - checker

@todo header
@todo based upon sge library

*/

#if !defined(LINALG_H)

// explicit dependencies
#include "ge.h"
#include <assert.h>


/*----------------------------------------------------------------------------

matrix - The simple and stupid matrix class.

*/

class matrix
{
public:
    matrix( int Rows, int Columns = 1 )
    {
        Initialize(Rows,Columns);
    }
    matrix( matrix const &m )
    {
        Initialize(m.GetRows(),m.GetColumns());
        for(int i = 0;i < m.GetRows();i++)
        {
            for(int j = 0;j < m.GetColumns();j++)
            {
                elem(Matrix,i,j) = m(i,j);
            }
        }
    }
    matrix &operator=( matrix const &m )
    {
        assert((m.GetRows() == GetRows()) && (m.GetColumns() == GetColumns()));
        for(int i = 0;i < m.GetRows();i++)
        {
            for(int j = 0;j < m.GetColumns();j++)
            {
                elem(Matrix,i,j) = m(i,j);
            }
        }
        return *this;
    }

    int GetRows( void ) const { return Matrix.rd; }
    int GetColumns( void ) const { return Matrix.cd; }

    real &operator()( int i, int j )
    {
        assert((i < Matrix.rd) && (j < Matrix.cd));
        return elem(Matrix,i,j);
    }
    real const &operator()( int i, int j ) const
    {
        assert((i < Matrix.rd) && (j < Matrix.cd));
        return elem(Matrix,i,j);
    }

    ~matrix( void )
    {
        Uninitialize();
    }        

private:

    void Initialize( int Rows, int Columns )
    {
        assert(Columns <= MAXCOL);
        Matrix.rd = Rows;
        Matrix.cd = Columns;
        for(int j = 0;j < Columns;j++)
        {
            Matrix.pd[j] = new float[Rows];
            assert(Matrix.pd[j]);
            for(int i = 0;i < Rows;i++)
            {
                Matrix.pd[j][i] = float(0);
            }
        }
    }

    void Uninitialize( void )
    {
        for(int j = 0;j < Matrix.cd;j++)
        {
            delete [] Matrix.pd[j];
        }
    }
        
    FULL Matrix;

    friend bool SolveLinearSystem( matrix const &A, matrix &x,
                                   matrix const &b );
};
    

/*----------------------------------------------------------------------------

functions

*/

matrix Transpose( matrix const &m )
{
    matrix mt(m.GetColumns(),m.GetRows());
    for(int i = 0;i < m.GetRows();i++)
    {
        for(int j = 0;j < m.GetColumns();j++)
        {
            mt(j,i) = m(i,j);
        }
    }
    return mt;
}

matrix operator-( matrix const &m )
{
    matrix mn(m.GetRows(),m.GetColumns());
    for(int i = 0;i < m.GetRows();i++)
    {
        for(int j = 0;j < m.GetColumns();j++)
        {
            mn(i,j) = -m(i,j);
        }
    }
    return mn;
}

matrix operator*( matrix const &a, matrix const &b )
{
    assert(a.GetColumns() == b.GetRows());
    matrix c(a.GetRows(),b.GetColumns());
    
    for(int i = 0;i < a.GetRows();i++)
    {
        for(int j = 0;j < b.GetColumns();j++)
        {
            for(int k = 0;k < a.GetColumns();k++)
            {
                c(i,j) += a(i,k) * b(k,j);
            }
        }
    }
    return c;
}
    
matrix operator+( matrix const &a, matrix const &b )
{
    assert(a.GetRows() == b.GetRows());
    assert(a.GetColumns() == b.GetColumns());
    matrix c(a.GetRows(),a.GetColumns());
    
    for(int i = 0;i < a.GetRows();i++)
    {
        for(int j = 0;j < a.GetColumns();j++)
        {
            c(i,j) = a(i,j) + b(i,j);
        }
    }
    return c;
}

matrix operator-( matrix const &a, matrix const &b )
{
    return a + (-b);
}

matrix operator*( real a, matrix const &b )
{
    matrix c(b.GetRows(),b.GetColumns());
    
    for(int i = 0;i < b.GetRows();i++)
    {
        for(int j = 0;j < b.GetColumns();j++)
        {
            c(i,j) = a * b(i,j);
        }
    }
    return c;
}

#if 1
#include <fstream.h>
void PrintMatrix( char const *pHeader, matrix const &m )
{
    ofstream Out("matrix.txt",ios::app);
    Out<<pHeader<<endl;
    for(int i = 0;i < m.GetRows();i++)
    {
        for(int j = 0;j < m.GetColumns();j++)
        {
            Out.width(5);
            Out.precision(3);
            Out<<m(i,j)<<" ";
        }
        Out<<endl;
    }
}
#endif

extern "C" int sgefa( FULL *m, int *pPivots );
extern "C" void sgesl( FULL *m, int *pPivots, float *b, int Job = 0 );
bool SolveLinearSystem( matrix const &A, matrix &x,
                        matrix const &b )
{
    assert(A.GetColumns() == x.GetRows());
    assert(x.GetRows() == b.GetRows());
    assert((x.GetColumns() == 1) && (b.GetColumns() == 1));

    bool ReturnValue = false;
    int aPivots[MAXCOL];

    matrix Af = A;
    if(!sgefa(&Af.Matrix,aPivots))
    {
        x = b;
        sgesl(&Af.Matrix,aPivots,x.Matrix.pd[0]);
        ReturnValue = true;
    }

    return ReturnValue;
}

#define LINALG_H
#endif
